"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const validator = require("validator");
const moment = require("moment");
const mongo_1 = require("./utils/mongo");
const params_1 = require("./params");
class ParameterValidator {
    static addRule(type, v, c) {
        ParameterValidator.typeMap[type] = c ? [v, c] : v;
    }
    static isPhone(str, options) {
        return /^1\d{10}$/.test(str);
    }
    static isDate(str, options) {
        try {
            return moment(str).isValid();
        }
        catch (err) {
            return false;
        }
    }
    static isString(str, options) {
        if (typeof str !== 'string') {
            return false;
        }
        return validator.isLength(str, options);
    }
    static isArray(obj, options) {
        const result = Array.isArray(obj);
        if (!result) {
            return false;
        }
        if (options.len) {
            return obj.length === options.len;
        }
        if (options.min) {
            return obj.length >= options.min;
        }
        if (options.max) {
            return obj.length <= options.max;
        }
        return true;
    }
    static isJSON(str, options) {
        const type = typeof str;
        if (type === 'string') {
            return validator.isJSON(str, options);
        }
        return type == 'object';
    }
    static isObjectId(str) {
        return validator.isLength(str, { min: 24, max: 24 });
    }
    static isContains(str, options) {
        return validator.isIn(str, options.values);
    }
    static parseDate(val) {
        if (val instanceof Date) {
            return val;
        }
        return moment(val).toDate();
    }
    static parseString(val) {
        return val.toString();
    }
    static parseBoolean(str) {
        return (typeof str === 'boolean') ? str : str === 'true';
    }
    static parseEnum(val, options) {
        if (!options.value) {
            return val;
        }
        const type = ParameterValidator.typeMap[options.value];
        if (type && type.converter) {
            return type.converter(val);
        }
        return val;
    }
    static parseJson(val, options) {
        if (typeof val === 'string') {
            try {
                return JSON.parse(val);
            }
            catch (err) {
                return {};
            }
        }
        return val;
    }
    static parseObjectId(val) {
        return mongo_1.default.toObjectId(val);
    }
    validateRules(rules, params, parent = '') {
        let errors = new Array();
        if (!rules || !params) {
            return errors;
        }
        for (const field in rules) {
            if (field === 'required') {
                continue;
            }
            const rule = rules[field];
            if (typeof rule !== 'object') {
                throw new Error('[' + field + ']不是有效的验证规则.');
            }
            let value = params[field];
            let required = (typeof rule.required !== 'boolean' || rule.required);
            const _field = parent ? [parent, field].join('.') : field;
            if (typeof rule.type !== 'string') {
                if (typeof value === 'undefined' && required) {
                    errors.push({ message: rule.message || '参数' + _field + '必须填写', field: _field });
                    continue;
                }
                const _errors = this.validateRules(rule, value, _field);
                if (_errors.length > 0) {
                    errors = errors.concat(_errors);
                }
                continue;
            }
            const valueType = typeof value;
            const hasValue = valueType === 'boolean' || valueType === 'number' || value;
            if (!hasValue) {
                if (!required) {
                    delete params[field];
                    continue;
                }
                errors.push({ message: rule.message || '参数' + _field + '必须填写', field: _field });
                continue;
            }
            let validatorFun = () => true;
            let convertFun = (val) => val;
            const type = ParameterValidator.typeMap[rule.type];
            if (!type) {
                throw new Error('验证规则[' + _field + ']不包含验证类型:' + rule.type);
            }
            validatorFun = type.validator || validatorFun;
            convertFun = type.converter || convertFun;
            value = type.character || typeof type.character === 'undefined' ? value.toString() : value;
            const result = validatorFun(value, rule);
            if (!result) {
                errors.push({ message: rule.message || '参数' + _field + '验证失败', field: _field });
                continue;
            }
            if (rule.type === params_1.ParamType.ARRAY && rule.schema) {
                const list = new Array();
                for (let index = 0; index < value.length; index++) {
                    const item = value[index];
                    let _result = undefined;
                    let _errors;
                    if (rule.schema && typeof rule.schema.type === 'string') {
                        const _value = { [field]: item };
                        _errors = this.validateRules({ [field]: rule.schema }, _value);
                        _result = _value[field];
                    }
                    else {
                        _errors = this.validateRules(rule.schema, item);
                        _result = item;
                    }
                    if (_errors.length > 0) {
                        errors = errors.concat(_errors);
                        break;
                    }
                    else if (typeof _result !== 'undefined') {
                        list.push(_result);
                    }
                }
                params[field] = list;
            }
            try {
                params[field] = convertFun(params[field], rule);
            }
            catch (err) {
                errors.push({ message: '参数' + field + '转换值失败。', field });
            }
        }
        return errors;
    }
    async validate(rules, params) {
        return this.validateRules(rules, params);
    }
}
ParameterValidator.typeMap = {
    [params_1.ParamType.NUMBER]: { validator: validator.isInt, converter: parseInt },
    [params_1.ParamType.INTEGER]: { validator: validator.isInt, converter: parseInt },
    [params_1.ParamType.FLOAT]: { validator: validator.isFloat, converter: parseFloat },
    [params_1.ParamType.STRING]: { validator: ParameterValidator.isString, converter: ParameterValidator.parseString, character: false },
    [params_1.ParamType.DATE]: { validator: ParameterValidator.isDate, converter: ParameterValidator.parseDate, character: false },
    [params_1.ParamType.BOOLEAN]: { validator: validator.isBoolean, converter: ParameterValidator.parseBoolean },
    [params_1.ParamType.ARRAY]: { validator: ParameterValidator.isArray, character: false },
    [params_1.ParamType.ENUM]: { validator: ParameterValidator.isContains, converter: ParameterValidator.parseEnum },
    [params_1.ParamType.EMAIL]: { validator: validator.isEmail },
    [params_1.ParamType.URL]: { validator: validator.isURL },
    [params_1.ParamType.HASH]: { validator: validator.isHash },
    [params_1.ParamType.JSON]: { validator: ParameterValidator.isJSON, converter: ParameterValidator.parseJson, character: false },
    [params_1.ParamType.JWT]: { validator: validator.isJWT },
    [params_1.ParamType.PHONE]: { validator: ParameterValidator.isPhone },
    [params_1.ParamType.OBJECT_ID]: { validator: ParameterValidator.isObjectId, converter: ParameterValidator.parseObjectId, }
};
exports.default = ParameterValidator;
