"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tsquery_1 = require("@phenomnomnominal/tsquery");
const TypeScript_1 = require("TypeScript");
const base_1 = require("./base");
var PropertyType;
(function (PropertyType) {
    PropertyType[PropertyType["String"] = 1] = "String";
    PropertyType[PropertyType["Number"] = 2] = "Number";
    PropertyType[PropertyType["Boolean"] = 3] = "Boolean";
    PropertyType[PropertyType["Date"] = 4] = "Date";
    PropertyType[PropertyType["Any"] = 5] = "Any";
    PropertyType[PropertyType["Json"] = 6] = "Json";
    PropertyType[PropertyType["ObjectId"] = 7] = "ObjectId";
    PropertyType[PropertyType["Schema"] = 8] = "Schema";
    PropertyType[PropertyType["Buffer"] = 9] = "Buffer";
})(PropertyType = exports.PropertyType || (exports.PropertyType = {}));
class ModelAstParser extends base_1.default {
    constructor() {
        super(...arguments);
        this._schemas = new Array();
    }
    get schemas() { return this._schemas; }
    parse() {
        const val = this.getReturnValue();
        this.name = val.name;
        return this.getModelSchema(val.id, val.name);
    }
    getSchema(name) {
        return this._schemas.find((p) => p.name === name);
    }
    addSchema(obj) {
        this._schemas.push(obj);
        return obj;
    }
    getModelNode(name) {
        const schema = this.getVariableNode(name);
        if (schema.length === 0) {
            throw new Error(`无法获得模型[${name}]的定义`);
        }
        const exp = tsquery_1.tsquery(schema[0], "NewExpression");
        if (exp.length === 0) {
            throw new Error(`模型[${name}]必须实例化。`);
        }
        const obj = tsquery_1.tsquery(exp[0], 'ObjectLiteralExpression');
        if (obj.length === 0) {
            throw new Error(`模型[${name}]必须定义对象结构。`);
        }
        return obj[0];
    }
    getModelSchema(nodeName, modelName) {
        try {
            modelName = this.getInterfaceName(modelName);
            const obj = this.getSchema(modelName);
            if (obj) {
                return obj;
            }
            const node = this.getModelNode(nodeName);
            const schema = this.visit(node, nodeName);
            if (schema) {
                schema._id = { type: PropertyType.ObjectId, value: 'ObjectId', memo: 'ID', required: true };
            }
            return this.addSchema({ name: modelName, schema });
        }
        catch (err) {
            return;
        }
    }
    getInterfaceName(name) {
        name = name.charAt(0).toUpperCase() + name.substring(1);
        if (!name.startsWith(this.name)) {
            name = this.name + name;
        }
        return name;
    }
    getPropertyType(node) {
        const any = { type: PropertyType.Any, text: 'any' };
        const arrayAny = { type: [PropertyType.Any], text: 'Array<any>' };
        switch (node.kind) {
            case TypeScript_1.SyntaxKind.Identifier: {
                const text = node.escapedText;
                switch (text) {
                    case 'String':
                        return { type: PropertyType.String, text: text.toLowerCase() };
                    case 'Number':
                        return { type: PropertyType.Number, text: text.toLowerCase() };
                    case 'Boolean':
                        return { type: PropertyType.Boolean, text: text.toLowerCase() };
                    case 'Date':
                        return { type: PropertyType.Date, text };
                    case 'Buffer':
                        return { type: PropertyType.Buffer, text };
                    case 'JSON':
                        return { type: PropertyType.Json, text: 'any' };
                    case 'Array':
                        return arrayAny;
                    default: {
                        const schema = this.getModelSchema(text, text);
                        if (schema) {
                            return { type: PropertyType.Schema, text: schema.name, name: schema.name };
                        }
                        return { type: PropertyType.Any, text };
                    }
                }
            }
            case TypeScript_1.SyntaxKind.PropertyAccessExpression: {
                const text = node.getFullText().trim();
                if (text.includes('ObjectId')) {
                    return { type: PropertyType.ObjectId, text: 'ObjectId' };
                }
                return { type: PropertyType.Any, text };
            }
            case TypeScript_1.SyntaxKind.ArrayLiteralExpression: {
                const value = this.getPropertyType(node.elements[0]);
                if (value && !Array.isArray(value.type)) {
                    return { type: [value.type], text: `${value.text}[]`, name: value.name };
                }
                return arrayAny;
            }
            default:
                return any;
        }
    }
    getProperty(node) {
        let value = this.getPropertyType(node.initializer);
        const memo = this.getTextValue(node.parent, 'memo') || '';
        const required = this.getBooleanValue(node.parent, 'required');
        const $type = this.getIdentifierText(node.parent, '$type');
        const $min = this.getNumberValue(node.parent, '$min');
        const $max = this.getNumberValue(node.parent, '$max');
        const $message = this.getTextValue(node.parent, '$message');
        const $keys = this.getIdentifierText(node.parent, '$keys');
        const $value = this.getIdentifierText(node.parent, '$value');
        const $enums = this.getIdentifierText(node.parent, '$enums');
        return {
            type: value.type,
            value: $keys ? $keys : value.text,
            name: value.name,
            memo,
            required,
            $type,
            $value,
            $min,
            $max,
            $message,
            $keys,
            $enums
        };
    }
    visit(parent, pName) {
        const data = {};
        for (const node of parent.properties) {
            const name = node.name.escapedText;
            if (node.initializer.kind === TypeScript_1.SyntaxKind.ObjectLiteralExpression) {
                data[name] = this.visit(node.initializer, name);
                continue;
            }
            if (name === 'type') {
                return this.getProperty(node);
            }
            else {
                const value = this.getPropertyType(node.initializer);
                const prop = {
                    type: value.type,
                    value: value.text,
                    memo: '',
                    required: true
                };
                data[name] = prop;
            }
        }
        return data;
    }
    getReturnValue() {
        const ret = this.query('ReturnStatement > CallExpression');
        if (ret.length === 0) {
            throw new Error('未定义返回值。');
        }
        if (tsquery_1.tsquery(ret[0], `Identifier[escapedText='model']`).length !== 1) {
            throw new Error('返回值非mongoose模型。');
        }
        const args = ret[0].arguments;
        if (args.length < 2) {
            throw new Error('调用参数数量应大于2个。');
        }
        const name = args[0].text;
        if (!name || !name.length) {
            throw new Error('调用参数必须指定模型名称');
        }
        const id = args[1].escapedText;
        if (!id || !id.length) {
            throw new Error('调用参数必须指定模型定义对象名称');
        }
        return { name, id };
    }
    getIdentifierNode(node, id) {
        return this.query(`PropertyAssignment > Identifier[escapedText='${id}']`, node);
    }
    getIdentifierText(node, id) {
        const obj = this.getIdentifierNode(node, id);
        if (obj.length === 0) {
            return;
        }
        const value = obj[0].parent.initializer;
        return value ? value.getText() : undefined;
    }
    getTextValue(node, id) {
        const obj = this.getIdentifierNode(node, id);
        if (obj.length === 0) {
            return;
        }
        const strNode = tsquery_1.tsquery(obj[0].parent, `StringLiteral`);
        return strNode.length ? strNode[0].text : undefined;
    }
    getBooleanValue(node, id) {
        const obj = this.getIdentifierNode(node, id);
        if (obj.length === 0) {
            return true;
        }
        const parent = obj[0].parent;
        return parent.initializer.kind === TypeScript_1.SyntaxKind.FalseKeyword ? false : true;
    }
    getNumberValue(node, id) {
        const obj = this.getIdentifierNode(node, id);
        if (obj.length === 0) {
            return;
        }
        const value = obj[0].parent.initializer;
        if (value && (value.kind === TypeScript_1.SyntaxKind.NumericLiteral || value.kind === TypeScript_1.SyntaxKind.PrefixUnaryExpression)) {
            return value.getText();
        }
    }
}
exports.default = ModelAstParser;
