"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const random_1 = require("../utils/random");
const moment = require("moment");
const iopath = require("path");
class AliyunStorage {
    constructor(host, client) {
        this.host = host;
        this.client = client;
    }
    async put(path, file) {
        let fileName;
        if (Buffer.isBuffer(file)) {
            fileName = iopath.basename(path);
        }
        else {
            const ext = iopath.extname(file).toLowerCase();
            fileName = moment().format('YYYYMMDDHHmmss') + random_1.default.generateNumber(5) + ext;
        }
        path = path + '/' + fileName;
        const result = await this.client.put(path, file);
        return {
            path,
            name: fileName,
            size: result.size,
            url: this.host + result.name
        };
    }
    async remove(path) {
        if (Array.isArray(path)) {
            await this.client.deleteMulti(path);
        }
        else {
            await this.client.delete(path);
        }
    }
}
exports.AliyunStorage = AliyunStorage;
class StorageService {
    static async init(storage, manager) {
        this.storage = storage;
        this.manager = manager;
    }
    static getSessionId(url) {
        const pos = url.lastIndexOf('?');
        const err = `路径${url}不存在会话ID参数`;
        if (pos === -1) {
            throw new Error(err);
        }
        const params = new URLSearchParams(url.substring(pos));
        const id = params.get(this.PARAM_ID);
        if (!id) {
            throw new Error(err);
        }
        return id;
    }
    static appendSessionId(url, sessionId) {
        return `${url}?${this.PARAM_ID}=${sessionId}`;
    }
    static async addFiles(files, path) {
        if (!files || files.length === 0) {
            throw new Error('未接收到上传的文件。');
        }
        const objs = new Array();
        if (Array.isArray(files)) {
            for (const file of files) {
                const obj = await this.storage.put(path, file.filepath);
                objs.push(obj);
            }
        }
        else {
            const obj = await this.storage.put(path, files);
            objs.push(obj);
        }
        const list = await this.manager.add(objs);
        for (const file of list) {
            file.url = this.appendSessionId(file.url, file.id);
        }
        return list;
    }
    static verfiy(urls) {
        if (!Array.isArray(urls)) {
            urls = [urls];
        }
        return urls.map((url) => this.getSessionId(url));
    }
    static async updateSession(urls, ref_id) {
        const list = this.verfiy(urls);
        await this.manager.updateRef(list, ref_id);
        return list.length;
    }
    static async removeFiles(ref_id) {
        const files = await this.manager.getRefFiles(ref_id);
        await this.storage.remove(files.map((file) => file.path));
        await this.manager.removeRef(ref_id);
        return files.length;
    }
    static async clearExpire() {
        const files = await this.manager.getExpire();
        if (files.length > this.MAX_ID) {
            throw new Error(`批量操作文件数不能大于${this.MAX_ID}`);
        }
        await this.storage.remove(files.map((file) => file.path));
        await this.manager.removeMany(files.map((file) => file.id));
        return files.length;
    }
}
StorageService.PARAM_ID = 'sid';
StorageService.MAX_ID = 1000;
exports.default = StorageService;
